#include "defo.h"


#include "find-symmetry.h"
#include "MyIntervalFunc.h"
#include "bodies.h"
#include "solutions-tests.h"
#include "test-and-help-functions.h"
#include "krawczyk-newton.h"
#include "searchalg.h"

#include "cc-nb.h"




#ifdef PLANAR
MyInterval scalarProduct(const MyInterval &xi,const  MyInterval &yi,const  MyInterval &xj,const  MyInterval &yj)
{
    return xi * xj + yi * yj ;
}
#endif
#ifdef SPATIAL
MyInterval scalarProduct(const MyInterval &x1,const MyInterval &y1,const MyInterval &z1,const MyInterval &x2,const MyInterval &y2,const MyInterval &z2)
{
  return x1*x2+y1*y2+z1*z2;
}
#endif


bool isUnique(Count zeroFound)
{
  if ((zeroFound.maybeZeros == 0)&& (zeroFound.zeros == 1))
    return true;
  cout << "Couldn't establish uniqueness!" << endl;
  return false;
}




void axialSymOX(const Bodies &bodies, Bodies &mirrorReflection)
/** the reflection with respect the X axis.  x->x, y->-y,z->-z */
{
    MyInterval tempY,tempZ;

    mirrorReflection = bodies;
    for (int i = 0; i < bodies.N - 1; i++)
    {
        tempY = -bodies.y(i);
        mirrorReflection.setY(i, tempY);
		#ifdef SPATIAL
		tempZ = -bodies.z(i);
        mirrorReflection.setZ(i, tempZ);
		#endif
    }
}

bool colinearCC(const Bodies &bodies)
/** if the CC is colinear - whether it is contained in the OX axis */
{
    for (int i = 0; i < bodies.N; i++)
    {
	    if (!isSingular(bodies.y(i)))
            return false; // a body has y_i != 0
	   #ifdef SPATIAL
        if (!isSingular(bodies.z(i)))
            return false; // a body  z_i != 0
	   #endif		
    }

    Bodies mirrorReflection(bodies);
    Bodies hullBodies(bodies);

    axialSymOX(bodies, mirrorReflection);
    intervalHullBodies(bodies, mirrorReflection, hullBodies);

    // the trash variables used for compatybility with the call of searchCC
    ofstream outFileTest;
	ofstream outFileFailed;  
    outFileTest.open(zerosTestedFile, ios::binary | ios::out | ios::trunc);
    outFileFailed.open(undecidedFile, ios::binary | ios::out | ios::trunc);
	// end of trash


    Count zeroFound = searchCC(hullBodies, false,outFileTest,outFileFailed); // testing stage
	      
    outFileTest.close();
    outFileFailed.close();                            

    if (isUnique(zeroFound)) return true;
    return blowUpAndProveUniqueSolution(hullBodies);
   
}


#ifdef PLANAR
MyInterval distFromZeroSq(MyInterval x, MyInterval y)
/** Returns the distance from zero, squared. */
{
    return (power(x, 2) + power(y, 2));
}
#endif
#ifdef SPATIAL
MyInterval distFromZeroSq(MyInterval x, MyInterval y, MyInterval z)
/** Returns the distance from zero, squared. */
{
    return (power(x, 2) + power(y, 2) + power(z, 2));
}
#endif





int findGroups(const Bodies &bodies, vector<vector<int>> &vec)
/** On return:
 * `vec[i]`: the vector containing bodies with equal distances from 0, i.e.:
 *     `vec[i][0] = i, vec[i][j] = k` => bodies `i` and `k` have equal distances from 0.
 * Returns the minimal size of the group != 1. If all groups are singletons,
 * returns 0.
 */
{
    int N = bodies.N;
    int minCard = N + 1;
    int card;

    // Create a vector of size N with all values set to 0.
    vector<int> markers(N, 0);

    for (int i = 0; i < N; i++)
    {
        MyInterval xi = bodies.x(i);
        MyInterval yi = bodies.y(i);
		#ifdef PLANAR
		MyInterval distXYZi = distFromZeroSq(xi, yi);
		#endif
		#ifdef SPATIAL
        MyInterval zi = bodies.z(i);
        MyInterval distXYZi = distFromZeroSq(xi, yi, zi);
        #endif
        vector<int> row;  // Create an empty row
        row.push_back(i); // Add an element (column) to the row

        card = 1;
        for (int j = 0; j < N; j++)
        {
            if (j == i)
                continue;

            MyInterval xj = bodies.x(j);
            MyInterval yj = bodies.y(j);
			#ifdef PLANAR
		    MyInterval distXYZj = distFromZeroSq(xj, yj);
		    #endif
		    #ifdef SPATIAL
            MyInterval zj = bodies.z(j);
            MyInterval distXYZj = distFromZeroSq(xj, yj, zj);
			#endif

            // q(i) and q(j) have the same distance to (0,0)
            if (!emptyIntersection(distXYZi, distXYZj))
            {

                card++;
                row.push_back(j); // Add an element (column) to the row
            }
        }
        vec.push_back(row); // Add the row to the main vector
        if (1 < card && card < minCard)
            minCard = card;
    }

    // all the groups are singletons
    if (minCard == N + 1)
        minCard = 0;
    return minCard;
}




bool symmetricXCC(const Bodies &bodies)
//  For equal masses case only
// Returns `true` if the CC is symmetrical with respect to the OX axis (in the planar case) and OXY plane in the spatial case
{
    if (normalizationMethod != NormalizationMethod::equalToZero)
    {
        cout << "Normalization method not supported by symmetricXCC!" << endl;
        exit(1);
    }

    int NumBodies = bodies.N;

    // we set permutation=identity
    int permutation[ NumBodies ];
    for (int i = 0; i <  NumBodies ; i++)
        permutation[i] = i;
    
	// we try to find a candidate for a possible permutation 
	//   which is in agreement with the reflection with repect to the OX axis
    for (int i = 0; i < NumBodies; i++)
    {
        // equalToZero normalization method:
        if (i == NumBodies - 2)
            continue;
        
		#ifdef PLANAR
        if (!isSingular(bodies.y(i))) // doesn't contain 0
		#endif
		#ifdef SPATIAL
		 if (!isSingular(bodies.y(i)) || !isSingular(bodies.z(i))) // doesn't contain 0
		#endif
        {
            MyInterval myi = -bodies.y(i);
            MyInterval xi = bodies.x(i);
			#ifdef SPATIAL
			MyInterval mzi = -bodies.z(i);
			#endif
            for (int j = 0; j < NumBodies; j++)
            {
                if (i == j)
                    continue;
                MyInterval xj = bodies.x(j);
                MyInterval yj = bodies.y(j);
				#ifdef SPATIAL
				MyInterval zj = bodies.z(j);
				#endif
				#ifdef PLANAR
                if (!emptyIntersection(xi, xj) && !emptyIntersection(myi, yj))
				#endif
				#ifdef SPATIAL
				if (!emptyIntersection(xi, xj) && !emptyIntersection(myi, yj) && !emptyIntersection(mzi, zj))
				#endif
                {
                    permutation[i] = j;
                    permutation[j] = i;
                }
            }
        }
    }
    // now permutation - contains a candidate for symmetry
	
	
	 // below we construct symBodies --- a set which contains interval hull of bodies and its symmetric image
    Bodies symBodies(bodies);

    MyInterval xi, xj;
    MyInterval yi, yj;
    #ifdef SPATIAL
	MyInterval zi, zj;
    #endif 
    for (int i = 0; i < NumBodies - 2; i++)
    {
        if (permutation[i] == i)
        {
            yi = bodies.y(i);
			#ifdef SPATIAL
			zi = bodies.z(i);
			#endif
			
			#ifdef PLANAR
            if (isSingular(yi)){ // if the body lies on the x axis
                symBodies.setY(i, intervalHull(yi, -yi));
            }
			#endif
			#ifdef SPATIAL
			if (isSingular(yi) && isSingular(zi)) // if the body lies on the x axis
            {
                symBodies.setY(i, intervalHull(yi, -yi));
                symBodies.setZ(i, intervalHull(zi, -zi));
            }
			#endif
			else
            {
                cout << "there is no symmetry with respect to the X axis" << endl;
                return false; // because yi doesn't have a "reflection"
            }
        }
        else
        {
            xi = bodies.x(i);
            xj = bodies.x(permutation[i]);
            yi = bodies.y(i);
            yj = bodies.y(permutation[i]);

            symBodies.setX(i, intervalHull(xi, xj));
            symBodies.setY(i, intervalHull(yi, -yj));
			#ifdef SPATIAL
			zi = bodies.z(i);
            zj = bodies.z(permutation[i]);

            symBodies.setZ(i, intervalHull(zi, -zj));
			#endif
        }
    }

    // the trash variables used for compatybility with the call of searchCC
    ofstream outFileTest;
	ofstream outFileFailed;  
    outFileTest.open(zerosTestedFile, ios::binary | ios::out | ios::trunc);
    outFileFailed.open(undecidedFile, ios::binary | ios::out | ios::trunc);
	// end of trash



    Count zeroFound = searchCC(symBodies, false,outFileTest,outFileFailed);  // testing stage
	
	outFileTest.close();
	outFileFailed.close();
	                                  
    if(isUnique(zeroFound))
	   return true;
    if(blowUpAndProveUniqueSolution(symBodies))
       return true;
    else
	{
      cout << "undecided case!" << endl;
      return false;
    }
    return false;
}




#ifdef PLANAR
#include "find-sym-2d.cpp"
#endif
#ifdef SPATIAL
#include "find-sym-3d.cpp"
#endif